# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2015 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##

import SimpleXMLRPCServer
import logging
import threading
import xmlrpclib

import stoq
from stoqlib.lib.configparser import get_config

import stoqserver

logger = logging.getLogger(__name__)


class _RequestHandler(SimpleXMLRPCServer.SimpleXMLRPCRequestHandler):
    # Keep compatibility with old rpc path
    rpc_paths = ('/XMLRPC', )


class XMLRPCServer(object):

    def __init__(self, pipe_conn):
        self._pipe_conn = pipe_conn

    #
    #  Methods
    #

    def ping(self):
        logger.info("XMLRPC: ping")
        return "pong"

    def version(self):
        logger.info("XMLRPC: version")
        return stoqserver.version_str

    def stoq_version(self):
        logger.info("XMLRPC: stoq_version")
        return stoq.version

    def restart(self):
        logger.info("XMLRPC: restart")
        t = threading.Timer(1.0, self._run_action, args=('restart', ))
        t.start()
        return "Restart command sent..."

    def get_backup_key(self):
        config = get_config()
        return config.get('Backup', 'key')

    def set_backup_key(self, key):
        config = get_config()
        config.set('Backup', 'key', key)
        config.flush()
        # Restart stoqserver so the backup key will take effect immediately
        self.restart()
        return "Backup key set successfully"

    def pause_tasks(self):
        return self._run_action('pause_tasks')

    def resume_tasks(self):
        return self._run_action('resume_tasks')

    def htsql_query(self, query):
        return self._run_action('htsql_query', query)

    def backup_database(self):
        return self._run_action('backup_database')

    def backup_status(self, user_hash=None):
        return self._run_action('backup_status', user_hash)

    def backup_restore(self, user_hash, time=None):
        return self._run_action('backup_restore', user_hash, time)

    def plugin_action(self, plugin_name, task_name, action, *args):
        return self._run_action(
            'plugin_action', plugin_name, task_name, action, args)

    def register_link(self, pin):
        return self._run_action('register_link', pin)

    def install_plugin(self, plugin_name):
        return self._run_action('install_plugin', plugin_name)

    #
    #  Private
    #

    def _run_action(self, action, *args):
        logger.info("XMLRPC: action %s(%s)",
                    action, ', '.join('"%s"' % (a, ) for a in args))
        self._pipe_conn.send((action, ) + args)
        retval, msg = self._pipe_conn.recv()
        if not retval:
            raise xmlrpclib.Fault(32000, msg)
        return msg


def run_xmlrpcserver(pipe_conn, port):
    server = SimpleXMLRPCServer.SimpleXMLRPCServer(
        ('', port), requestHandler=_RequestHandler, allow_none=True)
    server.register_introspection_functions()
    server.register_instance(XMLRPCServer(pipe_conn))
    server.serve_forever()



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2015 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##

import collections
import datetime
import logging
import multiprocessing
import os
import random
import shutil
import signal
import subprocess
import sys
import tempfile
import time

import dateutil.parser
from stoqlib.api import api
from stoqlib.exceptions import DatabaseError
from stoqlib.database.runtime import get_default_store, set_default_store
from stoqlib.database.settings import db_settings
from stoqlib.domain.plugin import PluginEgg
from stoqlib.lib.configparser import get_config
from stoqlib.lib.pluginmanager import get_plugin_manager
from stoqlib.lib.settings import UserSettings

from stoqserver import library
from stoqserver.common import APP_BACKUP_DIR, SERVER_XMLRPC_PORT
from stoqserver.lib import backup
from stoqserver.lib.xmlrpcresource import run_xmlrpcserver
from stoqserver.server import StoqServer

logger = logging.getLogger(__name__)
# Indicates if we are doing a backup. This uses a piece of shared memory
# so forked processes will all see the same value when updated
_doing_backup = multiprocessing.Value('i', 0)


class TaskException(Exception):
    """Default tasks exception"""


def _setup_signal_termination():
    def _sigterm_handler(_signal, _stack_frame):
        os._exit(0)
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, _sigterm_handler)


def backup_database(full=False):
    if not os.path.exists(APP_BACKUP_DIR):
        os.makedirs(APP_BACKUP_DIR)

    filename = os.path.join(APP_BACKUP_DIR, 'stoq.dump')
    if not db_settings.dump_database(filename, format='plain'):
        raise TaskException("Failed to dump the database")

    backup.backup(APP_BACKUP_DIR, full=full)
    logger.info("Database backup finished sucessfully")


def restore_database(user_hash, time=None):
    assert user_hash

    # If the database doesn't exist, get_default_store will fail
    try:
        default_store = get_default_store()
    except Exception:
        default_store = None

    if default_store is not None and db_settings.has_database():
        try:
            default_store.lock_database()
        except DatabaseError:
            raise TaskException(
                "Could not lock database. This means that there are other "
                "clients connected. Make sure to close every Stoq client "
                "before updating the database")
        except Exception:
            raise TaskException(
                "Database is empty or in a corrupted state. Fix or drop it "
                "before trying to proceed with the restore")
        else:
            default_store.unlock_database()

        with tempfile.NamedTemporaryFile() as f:
            if not db_settings.dump_database(f.name):
                raise TaskException("Failed to dump the database")
            backup_name = db_settings.restore_database(f.name)
            logger.info("Created a backup of the current database state on %s",
                        backup_name)

    tmp_path = tempfile.mkdtemp()
    try:
        # None will make the default store be closed, which we need
        # to sucessfully restore the database
        set_default_store(None)
        restore_path = os.path.join(tmp_path, 'stoq')

        backup.restore(restore_path, user_hash, time=time)

        db_settings.clean_database(db_settings.dbname, force=True)
        db_settings.execute_sql(os.path.join(restore_path, 'stoq.dump'),
                                lock_database=True)

        logger.info("Backup restore finished sucessfully")
    finally:
        # get_default_store will recreate it (since we closed it above)
        get_default_store()
        shutil.rmtree(tmp_path, ignore_errors=True)


def backup_status(user_hash=None):
    backup.status(user_hash=user_hash)


def start_xmlrpc_server(pipe_conn):
    _setup_signal_termination()
    logger.info("Starting the xmlrpc server")

    config = get_config()
    port = int(config.get('General', 'serverport') or SERVER_XMLRPC_PORT)

    run_xmlrpcserver(pipe_conn, port)


def start_server():
    _setup_signal_termination()
    logger.info("Starting stoq server")

    stoq_server = StoqServer()
    stoq_server.run()


def start_rtc():
    if not api.sysparam.get_bool('ONLINE_SERVICES'):
        logger.info("ONLINE_SERVICES not enabled. Not starting rtc...")
        return

    logger.info("Starting webRTC")

    cwd = library.get_resource_filename('stoqserver', 'webrtc')
    retry = True

    config = get_config()
    extra_args = []
    camera_urls = config.get('Camera', 'url') or None
    if camera_urls:
        extra_args.append('-c=' + ' '.join(set(camera_urls.split(' '))))

    xmlrpc_host = config.get('General', 'serveraddress') or '127.0.0.1'
    extra_args.append('-h={}'.format(xmlrpc_host))

    xmlrpc_port = config.get('General', 'serverport') or SERVER_XMLRPC_PORT
    extra_args.append('-p={}'.format(xmlrpc_port))

    while retry:
        retry = False
        popen = subprocess.Popen(
            ['bash', 'start.sh'] + extra_args, cwd=cwd)

        def _sigterm_handler(_signal, _stack_frame):
            popen.poll()
            if popen.returncode is None:
                popen.terminate()
            os._exit(0)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, _sigterm_handler)

        popen.wait()
        if popen.returncode == 11:
            logger.warning("libstdc++ too old, not running webRTC client. "
                           "A system upgrade may be required!")
            retry = False
        elif popen.returncode == 10:
            logger.warning("Something failed when trying to start webrtc. "
                           "Retrying again in 10 minutes...")
            time.sleep(10 * 60)
            retry = True
        elif popen.returncode == 12:
            logger.warning("webrtc installation corrupted. Restarting it...")
            time.sleep(1)
            retry = True
        elif popen.returncode == 139:
            logger.warning("Segmentation fault caught on wrtc. Restarting...")
            time.sleep(1)
            retry = True


def start_plugins_update_scheduler(event):
    _setup_signal_termination()

    if not api.sysparam.get_bool('ONLINE_SERVICES'):
        logger.info("ONLINE_SERVICES not enabled. Not scheduling plugin updates...")
        return

    manager = get_plugin_manager()

    while True:
        last_check_str = UserSettings().get('last-plugins-update', None)
        last_check = (dateutil.parser.parse(last_check_str)
                      if last_check_str else datetime.datetime.min)

        # Check for updates once per day
        if last_check.date() == datetime.date.today():
            time.sleep(24 * 60 * 60)
            continue

        logger.info("Checking for plugins updates...")
        updated = False
        default_store = get_default_store()
        # TODO: atm we are only updating the conector plugin to avoid
        # problems with migrations. Let this update everything once we
        # find a solution for this problem
        for egg in default_store.find(PluginEgg, plugin_name=u'conector'):
            md5sum = egg.egg_md5sum
            manager.download_plugin(egg.plugin_name)

            # If download_plugin updated the plugin, autoreload will
            # make egg. be reloaded from the database
            if md5sum != egg.egg_md5sum:
                updated = True

        settings = UserSettings()
        settings.set('last-plugins-update',
                     datetime.datetime.now().isoformat())
        settings.flush()

        if updated:
            # Wait until any running backup is finished and restart
            while _doing_backup.value:
                time.sleep(60)

            logger.info("Some plugins were updated. Restarting now "
                        "to reflect the changes...")
            event.set()
        else:
            logger.info("No update was found...")


def start_backup_scheduler():
    global _doing_backup
    _setup_signal_termination()

    if not api.sysparam.get_bool('ONLINE_SERVICES'):
        logger.info("ONLINE_SERVICES not enabled. Not scheduling backups...")
        return

    logger.info("Starting backup scheduler")

    config = get_config()
    backup_schedule = config.get('Backup', 'schedule')
    if backup_schedule is None:
        # By defualt, we will do 2 backups. One in a random time between
        # 9-11 or 14-17 and another one 12 hours after that.
        # We are using 2, 3 and 4 because they will be summed with 12 bellow
        hour = random.choice([2, 3, 4, 9, 10])
        minute = random.randint(0, 59)
        backup_schedule = '%d:%d,%s:%d' % (hour, minute, hour + 12, minute)
        config.set('Backup', 'schedule', backup_schedule)
        config.flush()

    backup_hours = [map(int, i.strip().split(':'))
                    for i in backup_schedule.split(',')]
    now = datetime.datetime.now()
    backup_dates = collections.deque(sorted(
        now.replace(hour=bh[0], minute=bh[1], second=0, microsecond=0)
        for bh in backup_hours))

    while True:
        now = datetime.datetime.now()
        next_date = datetime.datetime.min
        while next_date < now:
            next_date = backup_dates.popleft()
            backup_dates.append(next_date + datetime.timedelta(1))

        time.sleep(max(1, (next_date - now).total_seconds()))

        for i in xrange(3):
            # FIXME: This is SO UGLY, we should be calling backup_database
            # task directly, but duplicity messes with multiprocessing in a
            # way that it will not work
            args = sys.argv[:]
            for i, arg in enumerate(args[:]):
                if arg == 'run':
                    args[i] = 'backup_database'
                    break

            _doing_backup.value = 1
            try:
                p = subprocess.Popen(args)
                stdout, stderr = p.communicate()
            finally:
                _doing_backup.value = 0

            if p.returncode == 0:
                break
            else:
                logger.warning(
                    "Failed to backup database:\nstdout: %s\nstderr: %s",
                    stdout, stderr)
                # Retry again with a exponential backoff
                time.sleep((60 * 2) ** (i + 1))



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2015 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##

import ConfigParser
import hashlib
import io
import os
import socket
import sys
import tempfile
import urllib2
import urlparse

import gtk
import netifaces
from zeroconf import ServiceBrowser, Zeroconf

from stoqserver.common import (APP_EGGS_DIR, SERVER_EGGS,
                               SERVER_EXECUTABLE_EGG, AVAHI_STYPE,
                               SERVER_XMLRPC_PORT)

_ = lambda s: s


class _StoqClient(gtk.Window):
    def __init__(self, *args, **kwargs):
        gtk.Window.__init__(self, *args, **kwargs)

        if not os.path.exists(APP_EGGS_DIR):
            os.makedirs(APP_EGGS_DIR)

        self._iters = {}

        self.executable_path = None
        self.conf_path = None
        self.python_paths = []

        self._setup_widgets()

    #
    #  Zeroconf Listener
    #

    def remove_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)
        # FIXME: How to remove the service when info is None?
        if info is None:
            return

        key = (info.address, info.port)
        if key in self._iters:
            self.store.remove(self._iters[key].pop())

    def add_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)

        server_address = 'http://%s:%s' % (
            socket.inet_ntoa(info.address), info.port)
        args = info.properties

        itr = self.store.append([server_address, args])
        self._iters[(info.address, info.port)] = itr

        if not self.selection.get_selected()[1]:
            self.selection.select_iter(itr)

    #
    #  Private
    #

    def _setup_widgets(self):
        vbox = gtk.VBox(spacing=6)

        self.store = gtk.ListStore(str, object)
        self.treeview = gtk.TreeView(self.store)

        self.server_column = gtk.TreeViewColumn(_("Server"))
        self.cell = gtk.CellRendererText()
        self.server_column.pack_start(self.cell, True)
        self.server_column.add_attribute(self.cell, 'text', 0)

        self.treeview.append_column(self.server_column)
        self.selection = self.treeview.get_selection()
        self.selection.connect('changed', self._on_treeview_selection__changed)
        vbox.pack_start(self.treeview, expand=True)

        self.username = gtk.Entry()
        username_hbox = gtk.HBox(spacing=6)
        username_hbox.pack_start(gtk.Label(_("Username:")), expand=False)
        self.username.connect('activate', self._on_username__activate)
        self.username.connect('changed', self._on_username__changed)
        username_hbox.pack_start(self.username, expand=True)
        vbox.pack_start(username_hbox, expand=False)

        self.password = gtk.Entry()
        self.password.set_property('visibility', False)
        password_hbox = gtk.HBox(spacing=6)
        password_hbox.pack_start(gtk.Label(_("Password:")), expand=False)
        self.password.connect('activate', self._on_password__activate)
        password_hbox.pack_start(self.password, expand=True)
        vbox.pack_start(password_hbox, expand=False)

        self.login_btn = gtk.Button(_("Start"))
        self.login_btn.connect('activate', self._on_login_btn__activate)
        vbox.pack_start(self.login_btn, expand=False)
        self.login_btn.set_sensitive(False)

        alignment = gtk.Alignment(0.5, 0.5, 1.0, 1.0)
        alignment.set_padding(6, 6, 6, 6)
        alignment.add(vbox)

        self.resize(400, 300)
        self.add(alignment)

        self.username.grab_focus()

    def _update_widgets(self):
        model, titer = self.selection.get_selected()
        self.login_btn.set_sensitive(
            bool(titer and self.username.get_text()))

    def _download_eggs(self, server_address, options):
        opener = self._get_opener(server_address)

        with io.BytesIO() as f:
            tmp = opener.open('%s/login' % (server_address, ))
            f.write(tmp.read())
            f.seek(0)
            config = ConfigParser.ConfigParser()
            config.readfp(f)

        parsed = urlparse.urlparse(server_address)
        address = parsed.netloc.split(':')[0]
        config.set('General', 'serveraddress', address)
        config.set('General', 'serverport', parsed.port or SERVER_XMLRPC_PORT)

        if not config.get('Database', 'address'):
            # If there's no database address, it means the server is running on
            # the same computer as the database. Set it on the config file, but
            # only if the client is not running on the same computer as the server
            for iface in netifaces.interfaces():
                addresses = [
                    i['addr'] for i in
                    netifaces.ifaddresses(iface).setdefault(
                        netifaces.AF_INET, [{'addr': None}])]
                if address in addresses:
                    break
            else:
                config.set('Database', 'address', address)

        with tempfile.NamedTemporaryFile(delete=False) as f:
            config.write(f)
            self.conf_path = f.name

        md5sums = {}
        tmp = opener.open('%s/md5sum' % (server_address, ))
        for line in tmp.read().split('\n'):
            if not line:
                continue
            egg, md5sum = line.split(':')
            md5sums[egg] = md5sum

        tmp.close()

        for egg in SERVER_EGGS:
            egg_path = os.path.join(APP_EGGS_DIR, egg)
            if not self._check_egg(egg_path, md5sums[egg]):
                with open(egg_path, 'wb') as f:
                    tmp = opener.open('%s/eggs/%s' % (server_address, egg))
                    f.write(tmp.read())
                    tmp.close()

            assert self._check_egg(egg_path, md5sums[egg])

            self.python_paths.append(egg_path)
            if egg == SERVER_EXECUTABLE_EGG:
                self.executable_path = egg_path

        return True

    def _get_opener(self, server_address):
        passman = urllib2.HTTPPasswordMgrWithDefaultRealm()
        passman.add_password(None, server_address,
                             self.username.get_text(),
                             hashlib.md5(self.password.get_text()).hexdigest())
        authhandler = urllib2.HTTPBasicAuthHandler(passman)
        return urllib2.build_opener(authhandler)

    def _check_egg(self, egg_path, md5sum):
        if not os.path.exists(egg_path):
            return False

        md5 = hashlib.md5()
        with open(egg_path, 'rb') as f:
            for chunk in iter(lambda: f.read(md5.block_size), b''):
                md5.update(chunk)

        return md5.hexdigest() == md5sum

    def _start(self):
        model, titer = self.treeview.get_selection().get_selected()
        if not titer:
            return

        if not self._download_eggs(
                model.get_value(titer, 0), model.get_value(titer, 1)):
            return

        self.hide()
        gtk.main_quit()

    #
    #  Callbacks
    #

    def _on_treeview_selection__changed(self, selection):
        self._update_widgets()

    def _on_username__changed(self, entry):
        self._update_widgets()

    def _on_username__activate(self, entry):
        self.password.grab_focus()

    def _on_password__activate(self, entry):
        self._start()

    def _on_login_btn__activate(self, button):
        self._start()


def main(args):
    try:
        # FIXME: Maybe we should not use zeroconf and instead implement
        # our own avahi browser.
        zeroconf = Zeroconf()
        client = _StoqClient()
        client.show_all()
        ServiceBrowser(zeroconf, '%s.local.' % (AVAHI_STYPE, ), client)
        gtk.gdk.threads_init()
        gtk.main()
    finally:
        zeroconf.close()

    env = os.environ.copy()
    env['PYTHONPATH'] = ':'.join(
        client.python_paths + [env.get('PYTHONPATH', '')])

    args = [sys.executable, client.executable_path, '-f', client.conf_path]
    os.execve(args[0], args, env)



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2015 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##

import os
from kiwi.dist import setup, listpackages, listfiles

import stoqserver


PACKAGE = 'stoqserver'

scripts = [
    'bin/stoqserver',
]
data_files = [
    (os.path.join(os.sep, 'etc', 'sudoers.d'),
     [os.path.join('data', 'sudoers.d', 'stoqserver')]),
    (os.path.join(os.sep, 'etc', 'supervisor', 'conf.d'),
     [os.path.join('data', 'supervisor', 'stoqserver.conf')]),
    (os.path.join(os.sep, 'usr', 'share', 'stoqserver', 'webrtc'),
     [os.path.join('data', 'webrtc', 'package.json'),
      os.path.join('data', 'webrtc', 'start.sh')] +
     listfiles('data', 'webrtc', '*.js')),
]

with open('requirements.txt') as f:
    install_requires = [l.strip() for l in f.readlines() if
                        l.strip() and not l.startswith('#')]

setup(
    name=PACKAGE,
    author="Stoq Team",
    author_email="stoq-devel@async.com.br",
    description="Stoq server",
    url="http://www.stoq.com.br",
    license="GNU LGPL 2.1 (see COPYING)",
    long_description=("Service that provides a bridge between Stoq and "
                      "stoq.link, along with other usefullnesses."),
    version=stoqserver.version_str + '.post1',
    packages=listpackages('stoqserver'),
    data_files=data_files,
    install_requires=install_requires,
    scripts=scripts,
    zip_safe=True,
)
